# -*- coding: utf-8 -*-
"""
Created on Thu Mar 10 22:05:56 2016

@author: Tobias Jachowski
"""
import collections
import numpy as np
import operator

from .region import Region
from .record import Record
from ..modification import Modification


class View(Region):
    """
    Defines a View on a given Region (Record or View), the parent of that view.
    The parent is an instance of a MultiRegion, which gives the options to add
    even further parents (see MultiRegion).
    The region of a View of the data to refer to can (in contrast to a Record)
    be choosen by setter methods (start, stop, tmin, tmax).
    """

    def __init__(self, parent=None, sticky_start=True, sticky_stop=True,
                 **kwargs):
        """
        Parameters
        ----------
        parent : Region
        sticky_start : bool
        sticky_stop : bool
        **kwargs

        Attributes
        ----------
        records
        record
        views
        modifications_apply
        modifications_based
        group
        parent
        calibration
        num_traces
        traces
        samplingrate
        start
        stop
        sticky_start : bool
            If True and `self.start` has the same position as the parent's
            start, `self.start` follows parent's start automatically, if it
            is changed.
            If False, only shrinkage of the parent's region (increasing start)
            updates `self.start`, if necessary.
            `self.sticky_start` is automatically set to False, if `self.start`
            is set manually.
        sticky_stop : bool
            If True and `self.stop` has the same position as the parent's
            stop, `self.stop` follows parent's stop automatically, if it
            is changed.
            If False, only shrinkage of the parent's region (decreasing stop)
            updates `self.stop`, if necessary.
            `self.sticky_stop` is automatically set to False, if `self.stop`
            is set manually.
        """
        super().__init__(**kwargs)

        if parent is None:
            raise TypeError("View missing 1 required positional argument: "
                            "`parent`")
        # Create a new MultiRegion with `parent` as parent and add it to `self`
        # as parent
        multiregion = MultiRegion(parent=parent, name=self.name)
        if not self.add_parent(multiregion):
            raise ValueError("Could not add `self` as a child, most probably "
                             "circular reference detected.")

        self.start = 0
        self.stop = self.parent.datapoints

        # Determine, whether start and stop (tmin and tmax) should follow the
        # increase of the parent region or compensate for it to point to the
        # same region.
        self.sticky_start = sticky_start
        self.sticky_stop = sticky_stop

    @property
    def records(self):
        return self.parent.records

    @property
    def record(self):
        return self.parent.record

    @property
    def views(self):
        return self.parent.views

    @property
    def modifications_apply(self):
        return self.modifications(key='apply')

    @property
    def modifications_based(self):
        return self.modifications(key='based')

    def modifications(self, key='apply'):
        if key == 'apply':
            return self.parent_instances(Modification)
        else:
            return self.child_instances(Modification)

    @property
    def group(self):
        modifications = list(self.modifications_based)
        if len(modifications) >= 1:
            # one or more modification_based -> same group as first
            # modification
            return modifications[0].group
        else:
            # no modification_based -> own group, which could be the same but
            # doesn't have to
            return self._group

    @group.setter
    def group(self, group):
        modifications = list(self.modifications_based)
        if len(modifications) >= 1:
            # one or more modification_based -> set group of first modification
            modifications[0].group = group
        else:
            # no modification_based -> set own group
            # self.group = None (or given group) is called in
            # super().__init__(), which ensures initialisation of self._group
            self._group = group

    @property
    def parent(self):
        # MultiRegion is the parent of View
        return self.first_ancestor_instance(MultiRegion, level=1, dft=False)

    @property
    def calibration(self):
        return self.parent.calibration

    @property
    def num_traces(self):
        if self.parent is None:
            return 0
        return self.parent.num_traces

    @property
    def traces(self):
        if self.parent is None:
            return []
        return self.parent.traces

    @property
    def samplingrate(self):
        if self.parent is None:
            return 0.0
        return self.parent.samplingrate

    @property
    def start(self):
        return self._start

    @property
    def stop(self):
        return self._stop

    @start.setter
    def start(self, start):
        if not hasattr(self, '_start'):
            # Initialise start, when called for the first time
            self._start = start
        else:
            if start != self._start:
                self.sticky_start = False
                # Validate start to be within 0 and self.parent.datapoints
                start = max(start, 0)
                start = min(start, self.parent.datapoints)

                # create shift for auto shift update of children
                shift = (self, 'start', 0, start - self._start)

                self._start = start

                # inform children - views and modifications based on this
                # view about change
                self.set_changed(level=1, index_shift=shift)

    @stop.setter
    def stop(self, stop):
        if not hasattr(self, '_stop'):
            # Initialise stop, when called for the first time
            self._stop = stop
        else:
            if stop != self._stop:
                self.sticky_stop = False
                # Validate stop to be within 0 and self.parent.datapoints
                stop = max(stop, 0)
                stop = min(stop, self.parent.datapoints)

                # create shift for auto shift update of children
                shift = (self,
                         'stop',
                         self._stop - self._start,
                         stop - self._stop)

                self._stop = stop

                # inform children - views and modifications based on this
                # view about change
                self.set_changed(level=1, index_shift=shift)

    def member_changed(self, ancestor=True, calledfromself=False,
                       index_shift=None, **kwargs):
        # If a change of an ancestor View or a MultiRegion was triggered by an
        # index_shift, check, if this index_shift of the ancestor necessitates
        # an update of self.stop or self.start. If yes, then set self.stop or
        # self.start accordingly and implicitly inform further descendants of
        # the change (The children of a View or a MultiRegion are informed upon
        # an index_shift, see call of set_changed() in property start/stop).
        # A change of descendants should be ignored.
        if index_shift is not None and not calledfromself and ancestor:
            if not self._update_index(index_shift):
                # It was not necessary to change start and stop, this means,
                # the data this View offers was not changed and in turn the
                # cached data does not have to be recalculated -> return and do
                # not trigger the recaching nor set the self.updated to False
                # (-> do not call super.member_changed())
                return

        # Set or unset cached data and update update status
        super().member_changed(ancestor=ancestor,
                               calledfromself=calledfromself, **kwargs)

    def _update_index(self, index_shift):
        """
        If the parent has changed its indexes, check, whether the own indexes
        need to be adjusted, too and inform child about index_shift, if
        necessary.

        Parameters
        ----------
        index_shift : tuple of (Region, str, int, int)
            The tuple contains: The caller Region that caused the index shift,
            the name of which index was changed (start/stop), the original
            index of the position where the change took place and the amount by
            which start or stop were shifted (caller, name, index, shift).

        Returns
        -------
        bool
            When `self.start` or `self.stop` had to be updated, return True,
            otherwise False.
        """
        # Get the start/stop shift from the parent
        caller, name, index, shift = index_shift
        # print("%s -> %s" %(caller.name, self.name))
        # print("%s, %s, %s" % (name, index, shift))

        # Initialize start and stop, in case there is now shift
        start = self._start
        stop = self._stop

        returnvalue = False

        if name == 'start':

            # There are three possible cases for self._start:
            # 1. start > index
            #    change start
            #    if shift > start - index:
            #      change start AND inform child
            # 2. start == index
            #    if shift is positive:
            #      leave start, inform child
            #    if shift is negative:
            #      if sticky: leave start, inform child
            #      else: change start
            # 3. start < index
            #    leave start, inform child
            #
            # There is one case, where self._stop is changed:
            # 1. stop > index
            #    change stop

            # Change self._start
            if self.sticky_start:
                op = operator.gt  # start > index is sticky
            else:
                op = operator.ge  # start >= index is non sticky
            if op(self._start, index):
                # Validate start to be at least as great as start of parent
                start = max(self._start - shift, index)

            # Change self._stop
            # and keep self._stop > index (start)
            if self._stop > index:
                # Validate stop to be at min as great as start + 1 of parent
                stop = max(self._stop - shift, index + 1)

            # Calculate index_shift for child
            _index = max(index - self._start, 0)
            _shift = max(shift - max(self._start - index, 0), 0)
            if self._start == index and self.sticky_start:
                _shift = shift

            # Set changed start and stop
            if self._start != start or self._stop != stop:
                # print(" Setting start in %s to %s" % (self.name, start))
                # print(" Setting stop in %s to %s" % (self.name, stop))
                # Set start and/or stop
                self._start = start
                self._stop = stop
                returnvalue = True

            # Inform child about 'start' index shift, if self._start and
            # self._stop do not compensate for the index shift
            if _shift != 0:
                i_shift = (self, name, _index, _shift)
                self.set_changed(level=1, index_shift=i_shift,
                                 includeself=False)

        else:  # name == 'stop':

            # Change self._start
            if self._start >= index:
                # This start follows the parent's stop index change
                start = self._start + shift
            else:  # self._start < self.parent.
                # Keep self._start < index (stop)
                # Validate start to be at max as great as stop - 1 of parent
                start = min(self._start, index + shift - 1)

            # Change self._stop
            if self.sticky_stop:
                op = operator.ge  # >= is sticky
            else:
                op = operator.gt  # > is non sticky
            if op(self._stop, index):
                # This start follows the parent's stop index change
                stop = self._stop + shift
            else:  # self._stop still could be after parent's start, therefore:
                stop = min(self._stop, index + shift)

            # Calculate index_shift for child
            _index = max(index - self._start, 0)
            _shift = min(shift + max(index - self._stop, 0), 0)
            if self._stop == index and self.sticky_stop:
                _shift = shift

            # Set changed start and stop
            if self._start != start or self._stop != stop:
                # print(" Setting start in %s to %s" % (self.name, start))
                # print(" Setting stop in %s to %s" % (self.name, stop))
                # Set start and/or stop
                self._start = start
                self._stop = stop
                returnvalue = True

            # Inform child about 'stop' index shift, if self._start and
            # self._stop do not (fully) compensate for the index shift
            if _index != 0 and _shift != 0:
                i_shift = (self, name, _index, _shift)
                self.set_changed(level=1, index_shift=i_shift,
                                 includeself=False)

        return returnvalue

    def _get_data_uncached(self, samples, traces_idx, copy):
        """
        Return uncached data, i.e.:
        Correct samples such that they relate to self.parent.datapoints, rather
        than self.datapoints, i.e. shift index by self.start.
        Get data from parent.
        Apply all modifications registered at this View.
        Return data.
        """

        # parentize the requested samples, i.e. correct for self.start
        if isinstance(samples, slice):
            p_start = samples.start + self.start
            p_stop = samples.stop + self.start
            p_step = samples.step
            p_samples = slice(p_start, p_stop, p_step)
        else:
            # samples is an np.array
            p_samples = samples + self.start

        data = self.parent._get_data(p_samples, traces_idx, copy)

        # modify data by applying all modifications to the unmodified data
        for mod in self.modifications():
            data = mod.modify(data, samples, traces_idx)

        return data


class MultiRegion(Region):
    """
    Holds references to multi parent instances (Region), offers methods
    to add and remove a parent and a method to serve the data of all parents
    concatenated. Inherits from class Region.
    """
    def __init__(self, parent=None, **kwargs):
        """
        Parameters
        ----------
        parent : Region or Iterable of Regions
        **kwargs

        Attributes
        ----------
        group
        start
        stop
        records
        views
        record
        calibration
        num_traces
        traces
        samplingrate
        """
        super().__init__(max_children=1, **kwargs)

        if parent is None:
            raise TypeError("Parent missing 1 required positional argument: "
                            "'parent'")

        if isinstance(parent, collections.Iterable):
            for p in parent:
                self.add_parent(p)
        else:
            self.add_parent(parent)

    def member_changed(self, ancestor=True, calledfromself=False,
                       index_shift=None, **kwargs):
        # Possible causes of an index shift:
        #    1. index shift of parent
        #    2. adding or removing a parent
        # This method is called in case 1. For handling of case 2 see
        # methods self.add_parent() and self.remove_parent().
        #
        # If a change of an ancestor View or a MultiRegion was triggered by an
        # index_shift, check, if this index_shift of the ancestor necessitates
        # an update of self.stop or self.start. If yes, then set self.stop or
        # self.start accordingly and implicitly inform further descendants of
        # the change (only the children of a View or a MultiRegion (level=1)
        # are informed upon an index_shift, see call of set_changed() in
        # property start/stop).
        # A change of descendants should be ignored.
        if index_shift is not None and not calledfromself and ancestor:
            caller, name, index, shift = index_shift
            # print("%s -> %s" %(caller.name, self.name))
            # print("%s, %s, %s" % (name, index, shift))

            # Find the start index of the changed caller
            start = 0
            for parent in self.parents:
                if parent is caller:
                    break
                start += parent.datapoints

            # Adjust index according to start index of parent/caller
            index = start + index

            # Create new index_shift based on old one
            index_shift = (self, name, index, shift)

            # Inform children - usually views - based on this MultiRegion about
            # index shift change
            self.set_changed(level=1, index_shift=index_shift,
                             includeself=False)

        # Set or unset cached data and update update status
        super().member_changed(ancestor=ancestor,
                               calledfromself=calledfromself, **kwargs)

    @property
    def group(self):
        try:
            return self.child.group
        except:
            return None

    @group.setter
    def group(self, group):
        try:
            self.child.group = group
        except:
            pass

    @property
    def start(self):
        return 0

    @property
    def stop(self):
        # the sum of all parent.datapoints
        datapoints = 0
        for parent in self.parents:
            datapoints += parent.datapoints
        return datapoints

    @property
    def records(self):
        """
        Return all parents of instance Record as a generator.
        """
        return self.parent_instances(Record)

    @property
    def views(self):
        """
        Return all parents of instance View as a generator.
        """
        return self.parent_instances(View)

    @property
    def record(self):
        """
        Returns the first record that view and subsequently that view's parents
        are based on.
        """
        # TODO: implement better decision, which record to return, depending on
        # the parents; this decision influences the properties: calibration,
        # samplingrate, and num_traces
        # Presently, only views/records can be added as parent, whose
        # calibration, samplingrate, and num_traces are equal to this view
        return self.first_ancestor_instance(Record)

    @property
    def calibration(self):
        # TODO: Views combine the raw signals -> no handling of different
        # calibration in View meaningful -> Create higher grouping system which
        # gives access to calibrated distances/extensions/forces, only, but not
        # the raw signal -> groups of Results? -> shift Calibration of Record
        # to Reults
        return self.record.calibration

    @property
    def num_traces(self):
        if self.record is None:
            return 0
        # TODO: Implemement handling of different num_traces: Allow only the
        # overlap of different traces to be used?
        return self.record.num_traces

    @property
    def traces(self):
        if self.record is None:
            return []
        return self.record.traces

    @property
    def samplingrate(self):
        if self.record is None:
            return 0.0
        # TODO: Implement handling of different samplingrate: up/down sampling?
        # Pandas? Every parent has its own samplingrate.
        return self.record.samplingrate

    def add_parent(self, region, index=None, after=None, before=None,
                   update_indices=True):
        """
        Priority of inserting the new region at a specific index in descending
        order:

            #. index
            #. after
            #. before

        Returns
        -------
        bool
            True if parent could be added or parent is already present in
            self._parents. Otherwise return False.
        """
        if region is None:
            return False

        # TODO: Implement possibility of handling different calibrations,
        # samplingrate, and num_traces
        # Check, if there is already at least one view existent and if so,
        # make sure the new parent is based on the same calibration, has
        # the same samplingrate and the same number of traces
        if (next(self.parents, None) is not None
            and (self.calibration is not region.calibration
                 or self.samplingrate != region.samplingrate
                 or self.num_traces != region.num_traces)):
            return False

        # Possible causes of an index shift:
        #    1. index shift of parent
        #    2. adding or removing a parent
        # This method is called in case 2. For handling of case 1 see method
        # self.member_changed().

        # Add parent.
        # Adding a parent usually causes a change of the children. Therefore,
        # super().add_parent() would call set_changed() itself, without any
        # parameters. This means, if a parent was added, set_changed() would
        # trigger a recalculation of Modification, regardless of whether the
        # change affected the region selected by the child View, or not.
        # However, adding a parent to a View only causes an index shift.
        # Therefore, disable generic set_changed() mechanism and initiate a
        # specific one with the index_shift information.
        if region in self.parents:
            # Region is already in parents.
            # Avoid informing the children about index_shift.
            # This is necessary, because further down super().add_parent()
            # would return True, and would therefore not prevent the
            # information about the index_shift.
            print("Region %s already referenced." % region)
            return True

        if not super().add_parent(region, index=index, after=after,
                                  before=before, set_changed=False):
            return False

        index_shift = None
        if update_indices:
            # Determine index, where insertion took place.
            index = 0
            for parent in self.parents:
                if parent is region:
                    break
                index += parent.datapoints
    
            # Create index_shift: (caller, name, index, shift)
            # Virtually, adding a region, is as if one would shift the stop of the
            # preceding region by region.datapoints.
            index_shift = (self, 'stop', index, region.datapoints)

        # Inform children about index_shift
        self.set_changed(level=1, index_shift=index_shift)

        return True

    def remove_parent(self, region, update_indices=True):
        if region is None:
            return False

        # Possible causes of an index shift:
        #    1. index shift of parent
        #    2. adding or removing a parent
        # This function is called in case 2. For handling of case 1 see method
        # self.member_changed().

        # Determine index, where removal will take place.
        index = 0
        for parent in self.parents:
            if parent is region:
                break
            index += parent.datapoints

        # Remove parent
        # super().remove_parent() would call set_changed() itself, without any
        # parameters. This  means, if a parent was removed, set_changed() would
        # trigger a recalculation of Modification, regardless of whether the
        # change affected the region selected by the child View, or not.
        # Therefore, disable generic set_changed() mechanism and initiate a
        # specific one with the index_shift information.
        if not super().remove_parent(region, set_changed=False):
            return False

        index_shift = None
        if update_indices:
            # Create index shift
            # Virtually, removing a region is as if one would shift the start of a
            # following region by region.datapoints.
            index_shift = (self, 'start', index, region.datapoints)

        # Inform children about index_shift,
        self.set_changed(level=1, index_shift=index_shift)

        return True

    def _get_data_uncached(self, samples, traces_idx, copy=True):
        """
        Return concatenated data from all parents.
        """

        # determine num_traces
        if isinstance(traces_idx, slice):
            stop = traces_idx.stop
            if stop < 0:
                stop = stop + len(self.traces)
            num_traces = int(np.ceil((stop - traces_idx.start)
                                     / traces_idx.step))
        else:
            # traces is an np.array
            num_traces = len(traces_idx)

        # determine the global boundaries of requested samples and datapoints
        if isinstance(samples, slice):
            s_start = samples.start
            s_stop = samples.stop
            s_step = samples.step
            datapoints = int(np.ceil((s_stop - s_start) / s_step))
        else:
            # samples is an np.array
            s_start = samples.min()
            s_stop = samples.max() + 1
            # datapoints = len(samples)
            # datapoints from every single parent need to be appended on the
            # fly (np.vstack)
            datapoints = 0

        # Go through all parents and get data from all parents, whose
        # datapoints fall within the requested samples region.
        # Initialize running indices start and stop, which keep track of
        # samples served by every single parent.
        start = 0
        stop = 0
        # initialize array, which will hold all data from parents
        data = np.empty((datapoints, num_traces))
        for parent in self.parents:
            # set lower bound of index of data served by current parent
            # -> start of current parent is stop of previous parent
            start = stop
            # set upper bound of index of data served by current parent
            # -> stop of current parent is:
            stop = start + parent.datapoints
            # Check, whether data of current parent is requested by samples
            if start < s_stop and stop > s_start:
                # create sub fraction of samples for current parent
                if isinstance(samples, slice):
                    # treat steps/decimate
                    if start <= s_start:
                        # requested data served by the first parent -> s_start
                        # is index of first datapoint -> no shift, correct
                        # sub_start by start
                        req_start = 0
                        sub_start = s_start - start
                    else:
                        # requested data served by a subsequent parent -> start
                        # is index of first datapoint -> shift start of data
                        # according to steps left over by num_samples % s_step
                        #  either some steps left over -> shift is ->
                        #   shift = s_step - steps_left_over
                        #  or if no stpes left over -> no shift ->
                        #   shift = s_step - s_step
                        shift = s_step - ((start - s_start) % s_step or s_step)
                        req_start = (start - s_start + shift) / s_step
                        sub_start = shift
                    req_stop = np.ceil((min(stop, s_stop) - s_start) / s_step)
                    sub_stop = min(stop, s_stop) - start
                    requested = slice(int(req_start), int(req_stop))
                    sub_samples = slice(sub_start, sub_stop, s_step)
                    # get data from parent
                    data[requested] = parent._get_data(sub_samples, traces_idx,
                                                       copy)
                else:
                    # samples is an np.array:
                    #   minimum index can be equal to start (0) or greater
                    #   maximum index has to be lesser than stop (datapoints)
                    requested = np.logical_and(samples >= start,
                                               samples < stop)
                    sub_samples = samples[requested] - start
                    # get data from parent and append to data
                    p_data = parent._get_data(sub_samples, traces_idx, copy)
                    data = np.vstack((data, p_data))

        return data



# -*- coding: utf-8 -*-
"""
Created on Sun Oct 16 22:03:33 2016

@author: Tobias Jachowski
"""
import numpy as np
import warnings

from . import signal as sn


class SignalFeature(object):
    """
    A `SignalFeature` analyzes a signal and detects specific sections in the
    signal (like steps, plateaus, etc.). It offers properties to easily get the
    indices/segments of the different sections. It provides the method
    `update()` to update the detected sections. `SignalFeature` is meant to be
    subclassed. The subclassed classes have to provide the function
    `_update()`, which will do the actual update of the sections.
    """
    def __init__(self, signals, resolution, **kwargs):
        self.updated = False
        self.update(signals, resolution)

    def update(self, signals, resolution):
        self._update(signals, resolution)
        self.updated = True

    def _update(self, signals, resolution):
        # do the update here
        pass


class StepFinder(SignalFeature):
    """
    A `StepFinder` finds the indices of positive and negative steps in a
    signal.
    """
    def __init__(self, signals, resolution, step_time=0.1, min_dwell_time=0.0,
                 include_first_last=False, **kwargs):
        """
        Parameters
        ----------
        signals : 2D numpy.ndarray
            Signal that contains steps to be detected.
        resolution : float
            Resolution of the signals in Hz.
        step_time : float
            The time a step takes in s. Should at least be greater as the
            greatest period that can be expected due to noise.
        min_dwell_time : float
            Min time between two successive steps in s. If dwell time between
            two detected steps is shorter the second step will be ignored.
        include_first_last : bool
            If `include_first_last` is True, the attributes `plateau_segments`,
            `pos_plateau_segments` and `neg_plateau_segments` will include the
            first datapoint to the first step and the last step to the last
            datapoint as a plateau.

        Attributes
        ----------
        include_first_last
        step_time
        min_dwell_time
        steps : 1D numpy.ndarray dtype=int
        pos_steps : 1D numpy.ndarray dtype=int
        neg_steps : 1D numpy.ndarray dtype=int
        plateau_segments : 2D numpy.ndarray dtype=int
            The first dimension contains the plateaus, the second dimension the
            start and stop value.
        pos_plateau_segments : 2D numpy.ndarray dtype=int
            The first dimension contains the plateaus, the second dimension the
            start and stop value.
        neg_plaeau_segments : 2D numpy.ndarray dtype=int
            The first dimension contains the plateaus, the second dimension the
            start and stop value.
        """
        self._step_time = step_time
        self._min_dwell_time = min_dwell_time
        self._include_first_last = include_first_last

        super().__init__(signals=signals, resolution=resolution)

    def _update(self, signals, resolution):
        signal = signals[:, 0]
        self.max_stop = signal.size
        self._steps = {}  # 'pos' and 'neg' steps of signal
        self._steps['pos'], self._steps['neg'] = sn.get_steps(
            signal, resolution, self.step_time, self.min_dwell_time)

    @property
    def include_first_last(self):
        return self._include_first_last

    @include_first_last.setter
    def include_first_last(self, include_first_last):
        self._include_first_last = include_first_last
        self.updated = False

    @property
    def step_time(self):
        """
        step_time : (s) time a step is supposed to take
            should at least be greater as the greatest period to be expected
            due to noise
        """
        return self._step_time

    @step_time.setter
    def step_time(self, step_time):
        self._step_time = step_time
        self.updated = False

    @property
    def min_dwell_time(self):
        """
        min_dwell_time : (s) min time between two successive steps
            if dwell time between two detected steps is shorter the second step
            will be ignored
        """
        return self._min_dwell_time

    @min_dwell_time.setter
    def min_dwell_time(self, min_dwell_time):
        self._min_dwell_time = min_dwell_time
        self.updated = False

    @property
    def steps(self):
        concatenated = np.r_[self._steps['pos'], self._steps['neg']]
        con_sort = np.sort(concatenated)
        return con_sort

    @property
    def pos_steps(self):
        return self._steps['pos']

    @property
    def neg_steps(self):
        return self._steps['neg']

    @property
    def plateau_segments(self):
        start = None
        stop = None
        if self.include_first_last:
            start = 0
            stop = self.max_stop
        return sn.idx_to_idx_segments(self.steps, start=start, stop=stop)

    @property
    def pos_plateau_segments(self):
        plateau_segments = self.plateau_segments
        idx = np.in1d(plateau_segments[:, 0], self.pos_steps)
        return plateau_segments[idx]

    @property
    def neg_plateau_segments(self):
        plateau_segments = self.plateau_segments
        idx = np.in1d(plateau_segments[:, 0], self.neg_steps)
        return plateau_segments[idx]


class CycleSectioner(SignalFeature):
    """
    A `CycleSectioner` analyses up to two signals for the excited signal ('x',
    'y'), the extrema ('minima', 'maxima'), the direction, i.e. positive
    or negative signal ('right', 'left'), and different cycles, i.e. rising
    ('rightstress', 'leftrelease') and falling ('rightrelease', 'leftstress')
    respectively.
    """
    def __init__(self, signals, resolution, compare_time=0.01,
                 threshold=5e-9, min_duration=0.2, highest_frequency=1,
                 reduce_false_extrema=False, **kwargs):
        """
        Parameters
        ----------
        signals : 2D numpy.ndarray
            One or two signals to be analysed.
        resolution : float
            Number of datapoints/s of the resolution of the signals.
        compare_time : float, optional
            Time in s for size of moving window used to compare amplitude of
            signals to each other.
        threshold : float, optional
            Value, the amplitude of a signal is compared to, if only one signal
            is provided.
        min_duration : float, optional
            Minimum duration in s a preliminary detected excited section has to
            have to be finally accepted as one.
        highest_frequency : float, optional
            Highest expected frequency in Hz of occurence of extrema in the
            signals. If you set a too low frequency, extrema will not be
            detected. A frequency to high will result in to many falsly
            positive detected extrema. See also parameter
            `reduce_false_extrema`.
        reduce_false_extrema : bool
            Set to True, to reduce falsly positive detected extrema. As a side
            effect, the extrema within half a wavelength of the given highest
            frequency at the ends of the signal could be shifted or even not be
            detected. Make sure to properly set `highest_frequency`, first.
            Another option to reduce False positives is to smoothen the
            signals, before.

        Attributes
        ----------
        excited : dict
            Use 'x' and 'y' keys to get excited segments for the corresponding
            signal axis.
        sections : dict
            Use 'x' and 'y keys for corresponding axis. This will give you a
            list of dicts, containing the extrema indices with the keys
            'minima' and 'maxima' and the section segments with the keys
            'rightstress', 'rightrelease', 'leftstress', 'leftrelease',
            'right', 'left', 'rise', and 'fall'.
        compare_time
        threshold
        min_duration
        highest_frequency
        reduce_false_extrema
        """
        self._compare_time = compare_time
        self._threshold = threshold
        self._min_duration = min_duration
        self._highest_frequency = highest_frequency
        self._reduce_false_extrema = reduce_false_extrema

        super().__init__(signals=signals, resolution=resolution)

    def _update(self, signals, resolution):
        # excited sections for axes 'x' and 'y'
        self.excited = {}
        # sections (minima, maxima, left, right, stress, release, rise, fall)
        # of excited sections for axes 'x' and 'y'
        self.sections = {}
        self._set_sections(signals, resolution)

    def _set_sections(self, signals, resolution):
        # Set excited sections
        self.excited['x'], self.excited['y'] \
            = sn.get_excited_sections(signals, resolution,
                                      threshold=self.threshold,
                                      compare_time=self.compare_time,
                                      min_duration=self.min_duration)[0]

        # Set sections with information of left, right, release, and stress,
        s = {}
        if signals.ndim >= 2 and signals.shape[1] >= 2:
            s['x'] = signals[:, 0]
            s['y'] = signals[:, 1]
        else:
            if signals.ndim == 1:
                s['x'] = signals
            else:
                s['x'] = signals[:, 0]
            s['y'] = np.empty((0, 2), dtype=int)
        for axis in ['x', 'y']:
            self.sections[axis] = []
            for section in self.excited[axis]:
                try:
                    signal = s[axis][section[0]:section[1]]
                    minima, maxima = sn.get_extrema(signal, resolution,
                                                    self.highest_frequency,
                                                    self.reduce_false_extrema)
                    posrise, posfall, negfall, negrise, pos, neg, rise, fall \
                        = sn.get_sections(signal, minima, maxima)[0]
                    shift = section[0]
                    seg_desc = {'minima': minima + shift,
                                'maxima': maxima + shift,
                                'rightstress': posrise + shift,
                                'rightrelease': posfall + shift,
                                'leftstress': negfall + shift,
                                'leftrelease': negrise + shift,
                                'right': pos + shift,
                                'left': neg + shift,
                                'rise': rise + shift,
                                'fall': fall + shift}
                    self.sections[axis].append(seg_desc)
                except:
                    warnings.warn("Could not determine the minima and maxima "
                                  "of the excited position.")

    @property
    def compare_time(self):
        return self._compare_time

    @compare_time.setter
    def compare_time(self, compare_time):
        self._compare_time = compare_time
        self.updated = False

    @property
    def threshold(self):
        return self._threshold

    @threshold.setter
    def threshold(self, threshold):
        self._threshold = threshold
        self.updated = False

    @property
    def min_duration(self):
        return self._min_duration

    @min_duration.setter
    def min_duration(self, min_duration):
        self._min_duratione = min_duration
        self.updated = False

    @property
    def highest_frequency(self):
        return self._highest_frequency

    @highest_frequency.setter
    def highest_frequency(self, highest_frequency):
        self._highest_frequency = highest_frequency
        self.updated = False

    @property
    def reduce_false_extrema(self):
        return self._reduce_false_extrema

    @reduce_false_extrema.setter
    def reduce_false_extrema(self, reduce_false_extrema):
        self._reduce_false_extrema = reduce_false_extrema
        self.updated = False



# -*- coding: utf-8 -*-
"""
Created on Wed Mar  9 17:12:08 2016

@author: Tobias Jachowski
"""
import importlib
import numbers
import numpy as np
import persistent

from . import calibsource as cs
from .. import config as cf
from .. import helpers as hp
from ..graph import GraphMember


class Calibration(GraphMember, persistent.Persistent):
    """
    A calibration gives you easy acces to the stiffness (kappa) and
    displacement sensitivity (beta) at the position, where the microsphere
    would touch the glass surface (intercept) and the dependency of kappa and
    beta of the position in Z (slope).

    A Calibration consists of:
      - a CalibrationSource, containing the distance dependent calibration
        factors
      - the positionZ dependent linear function of stiffness and displacement
        sensitivity
    """
    def __init__(self, calibsource, **kwargs):
        super().__init__(max_parents=0, **kwargs)

        if calibsource is None:
            raise TypeError("Calibration missing the required positional "
                            "argument 'calibsource'.")
        self.calibsource = calibsource

        # If the true distance of the position to the bead changed from the
        # measurement to the calibration (by dh), due to thermal expansion of
        # the stage/objective, the touchdown determined by the touchdown
        # modification corresponds to the dsurf determined by the calibration
        # script and therefore, no offset modification to the linear function
        # has to be applied:
        # h_stage - h_surf - dh != h_stage - h_surf - touchdown
        # If, however, there was no thermal expansion, then the linear function
        # (the intercept) has to be corrected for the touchdown offset:
        # beta0 = beta0 + mbeta*touchdown; kappa0 = kappa0 + mkappa*touchdown
        # To check, whether thermal expansion took place, one would need to
        # save the sum signal of the QPD during calibration and compare/map it
        # to the one saved during the measurement.Percent error of accidently
        # neglected
        # touchdown offset modification for the intercept:
        # offset != touchdown
        # self.offset = offset  # um, in case dsurf was not determined
        # correctly: offset of self.calibsource.dsurf, is set according to
        # touchdown of class Touchdown
        # self.correct_offset = True  # formerly thermal_expansion !=
        # correct_offset
        # correct the radius or assume the specified radius to be the real one?
        self.correct_radius = True

    def beta(self, positionZ=0.0):
        """
        Return the positionZ dependent displacement sensitivity.
        """
        return self._beta[0] + positionZ * self._beta[1]

    def kappa(self, positionZ=0.0):
        """
        Return the positionZ dependent stiffness.
        """
        return self._kappa[0] + positionZ * self._kappa[1]

    @property
    def _beta(self):
        """
        Give the slope (m) and intercept (y_0) of the positionZ dependent
        displacement sensitivity.
        See also intercept and slope for explanation.
        In (m/V)
        """
        return np.array([self.intercept('beta'), self.slope('beta')])

    @property
    def _kappa(self):
        """
        Give the slope (m) and intercept (y_0) of the positionZ dependent
        stiffness.
        See also intercept and slope for explanation.
        In (N/m)
        """
        return np.array([self.intercept('kappa'), self.slope('kappa')])

    def intercept(self, para):
        """
        Get the displacement sensitivity (para='beta') or stiffness
        (para='kappa') at the position, where the bead surface would touch the
        glass surface (the intercept). The position and in turn the value
        depends on the radius of the bead and therefore, whether the corrected
        radius or the specified radius is used for the calibration (choosen by
        attribute self.correct_radius).
        """
        # Give the intercept `y_0` for the function of 'surface distance' and
        # 'bead radius offset' corrected 'positionZ' (um)
        # (bead-surface glass-surface) for displacement sensitivity
        # (para='beta') and stiffness (para='kappa').
        # f(h_stage - dsurf - radius/focalshift) = y_0 + m*d
        if para == 'beta':
            return (self.calibsource.beta + self.calibsource.mbeta
                    * self.radius)
        elif para == 'kappa':
            return (self.calibsource.kappa + self.calibsource.mkappa
                    * self.radius)

    def slope(self, para):
        """
        Get the positionZ dependent slope of the displacement sensitivity
        (para='beta') or stiffness (para='kappa').
        """
        # Gives the slope `m` for the function of 'surface distance' and
        # 'bead radius offset' corrected 'positionZ' (um)
        # (bead-surface to glass-surface) for displacement sensitivity
        # (para='beta') and stiffness (para='kappa').
        # f(h_stage - dsurf - radius/focalshift) = y_0 + m*d
        # positionZ ~ - distance of calibsource -> negative slope
        if para == 'beta':
            return - (self.calibsource.mbeta * self.calibsource.focalshift)
        elif para == 'kappa':
            return - (self.calibsource.mkappa * self.calibsource.focalshift)

    def displacement(self, psd, positionZ=0.0, calib_traces=None):
        """
        Calculate the displacement.

        The displacement of psd is calculated with the displacement
        sensitivity stored in self.beta(). Height dependency of the
        displacement sensitivity is accounted for by supplying a positionZ
        signal.

        Parameters
        ----------
        psd : 2D array
            A 2D array containing the psd data.
        positionZ : int or array, optional
            Integer or array containing the height where each value of psd was
            measured. If an array is supplied, it needs to have the same length
            as psd.
        calib_traces : int, list, or slice, optional
            If self.beta() has another length or another order than the second
            dimension of psd, than you can specify which psd traces correspond
            to which self.beta() values. calib_traces needs to have as many
            values, as psd.shape[1].

        Returns
        -------
        2D array
            An array with the same dimensions as psd with displacement values.
        """
        if calib_traces is None:
            calib_traces = slice(0, psd.shape[1])
        if isinstance(positionZ, numbers.Number):
            positionZ = np.array([[positionZ]])
        if positionZ.ndim < 2:
            positionZ = positionZ[:, np.newaxis]
        beta = self.beta(positionZ=positionZ)[:, calib_traces]
        displacement = beta * psd
        return displacement

    def force(self, displacement, positionZ=0.0, calib_traces=None):
        """
        Calculate the force.

        The force of displacement is calculated with the stiffness stored in
        self.kappa(). Height dependency of the stiffness is accounted for by
        supplying a positionZ signal.

        Parameters
        ----------
        displacement : 2D array
            A 2D array containing the displacement (see self.displacement()).
        positionZ : int or array
            Integer or array containing the height where each value of
            displacement was measured. If an array is supplied, it needs to
            have the same length as psd.
        calib_traces : int, list, or slice
            If self.kappa() has another length or another order than the second
            dimension of displacement, than you can specify which displacement
            traces correspond to which self.kappa() values. calib_traces needs
            to have as many values, as psd.shape[1].

        Returns
        -------
        2d array
            An array with the same dimensions as displacement with force
            values.
        """
        if calib_traces is None:
            calib_traces = slice(0, displacement.shape[1])
        if isinstance(positionZ, numbers.Number):
            positionZ = np.array([[positionZ]])
        if positionZ.ndim < 2:
            positionZ = positionZ[:, np.newaxis]
        kappa = self.kappa(positionZ=positionZ)[:, calib_traces]
        force = kappa * displacement
        return force

    @property
    def dsurf(self):
        """
        Return the distance, where the bead center would have been on the glass
        surface during the calibration measurement.
        """
        return self.calibsource.dsurf

    @property
    def psurf(self):
        """
        Return the positionZ, were the bead center would have been on the glass
        surface during the calibration measurement.
        """
        return - self.dsurf

    @property
    def touchdown(self):
        """
        Return the positionZ, where the bead surface would have touched the
        glass surface during the calibration measurement.
        Uses either the corrected or specified radius of the bead determined by
        attribute `self.correct_radius`.
        """
        # Keep in mind, dsurf from the calibration script relates to distance,
        # which is ~ to - positionZ
        # -> !-! (dsurf !+! self.radius/self.focalshift)
        # = psurf - self.radius/self.focalshift
        return self.psurf - self.radius / self.focalshift

    @property
    def correct_radius(self):
        return self._correct_radius

    @correct_radius.setter
    def correct_radius(self, value):
        if not hasattr(self, '_correct_radius') \
                or value != self._correct_radius:
            self._correct_radius = value
            self.set_changed()

    @property
    def radius(self):
        if self.correct_radius:
            radius = self.calibsource.radiusspec * self.calibsource.corrfactor
        else:
            radius = self.calibsource.radiusspec
        return radius

    @property
    def focalshift(self):
        return self.calibsource.focalshift


def create_calibration(source_type=None, source_module=None, source_class=None,
                       cfgfile=None, verbose=False, name=None, **kwargs):
    """
    Parameters
    ----------
    source_type : str, optional
        Choose between different types of calibrationsources to be created or
        loaded. If None is selected and module or source_class are not given, a
        generic calibrationsource is created. The calibration factors are
        choosen in such a way that modules relying on a calibration are still
        (somewhat) functional.
        If a calibrationsource type is choosen, which needs to read in a file,
        additionally the parameters 'directory' and 'filename' have to be
        provided.
    source_module : str, optional
        A str of the name of a module a source_class is located in.
    source_class : str, optional
        A str of a class tha should be used as a CalibrationSource. The class
        has to be located in the module described.
    **kwargs : optional
        Any parameters, a CalibrationSource needs to be initialized.
    """
    # Set default configfile
    cfgfile = cfgfile or 'calibration.cfg'

    # Read configfile
    cfg = cf.read_cfg_file(cfgfile)

    # If directory and/or filename are given, join them to full path filename
    filename = kwargs.pop('filename', None)
    directory = kwargs.pop('directory', None)
    if filename is not None:
        _f, _d, filename = hp.file_and_dir(filename, directory)
        kwargs['filename'] = filename

    # Create a cs_class. If no source_module and source_class are given
    # by user, default to configfile. If configfile doesn't have entries
    # for source_module and source_class, default to std_mod and std_cls
    if source_module is None or source_class is None:
        # Get calibsource class
        cs_class = source_class or cf.get_cfg_class(
            cfg,
            sec=source_type,
            mod_opt='source_module',
            cls_opt='source_class',
            std_mod='.calibration',
            std_cls='CalibrationSource',
            verbose=verbose)
    else:
        mod = importlib.import_module(source_module, __package__)
        cs_class = getattr(mod, source_class)

    # Get parameters for cs_class
    _kwargs = cf.get_cfg_sec_dict(cfg, sec=source_type, verbose=verbose)

    # Remove options for cs_class creation (module and class)
    _kwargs.pop('source_module', None)
    _kwargs.pop('source_class', None)

    # Convert all comma separated strings into list of floats
    for key, value in _kwargs.items():
        if isinstance(value, str):
            _kwargs[key] = [float(l.strip()) for l in value.split(',')]

    # update options for cs_class, user given kwargs take precedence over
    # options retrieved from configfile
    _kwargs.update(kwargs)

    # If it is a standard calibration, put the source_type into the name.
    cs_name = None
    if cs_class is cs.CalibrationSource and source_type is not None:
        cs_name = ''.join(('Standard calibration source: ', source_type))

    # Create calibrationsource
    calibsource = cs_class(name=cs_name, **_kwargs)

    # remove parameters used by calibsource
    for par in ['beta', 'kappa', 'radiusspec', 'focalshift',
                'mbeta', 'mkappa', 'corrfactor', 'dsurf']:
        _kwargs.pop(par, None)

    # create calibration
    return Calibration(calibsource, name=name, **_kwargs)



# -*- coding: utf-8 -*-
"""
Created on Wed Mar 30 11:55:34 2016

@author: Tobias Jachowski
"""
import numpy as np
import matplotlib.pyplot as plt

from . import signal as sn
from .. import gui
from .signalfeature import SignalFeature


class Evaluator(object):
    """
    An Evaluator extracts information from data provided by a Region and
    provides functions to further process and evaluate the data.

    The Evaluator holds references to one Region and one SignalFeature. The
    Region provides the data (amongst other specific details) to be evaluated.
    The SignalFeature analyses a signal from the data and identifies specific
    sections in the signal (like steps, plateaus, etc.). The Evaluator uses
    these identified sections to extract further information and processed data
    from the data provided by the Region.

    The Evaluator provides convenient functions to adjust the resolution of the
    (extracted) data and identified sections it returns.
    """
    def __init__(self, region=None, calibration=None, resolution=None,
                 filter_time=-1.0, sf_class=None, traces_sf=None,
                 resolution_sf=None, filter_time_sf=None, **kwargs):
        """
        Parameters
        ----------
        region
        calibration : pyoti.calibration.calibration.Calibration, optional
        resolution : float, optional
            Desired resolution of data (and sections) returned by this
            Evaluator in datapoints per s. The resolution is automatically
            limited to be between 1 and `region.samplingrate` datapoints per s.
            Additionally, resolution is limited to values of
            `self.samplingrate` / decimate, where decimate is a positive
            integer. The actual resolution can be checked with the attribute
            `self.resolution'.The resolution defaults to `region.samplingrate`.
        filter_time : float, optional
            Desired time containing datapoints to be used with a running
            filter. If `filter_time` == 0.0, averaging is switched off. If
            `filter_time` > 0.0, signal is filtered according to the given
            `filter_time`. If `filter_time` < 0.0 (default), signal is filtered
            according to `resolution` (`self.decimate`). The actual datapoints
            beeing used to filter the data can be checked with the property
            `self.window`.
        sf_class : pyoti.evaluate.signalfeature.SignalFeature
            The class of the SignalFeature to be used for signal analysis.
        traces_sf : str, int, list of str/int, slice
            The traces of the signal, the SignalFeature needs to analyze.
        resolution_sf : float, optional
            The resolution the SignalFeature should work with.
            Defaults to `resolution` (or `region.samplingrate`). Actual
            reslution can be checked with the attribute `self.resolution_sf`.
        filter_time_sf : float, optional
            The time for the moving filter, the signal for the SignalFeature
            should be filtered with. See parameter `filter_time` for details.
            Defaults to `filter_time`. The actual datapoints beeing used to
            filter the signal for the SignalFeature can be checked with the
            property `self.window_sf`.

        Attributes
        ----------
        traces_sf : str, int, list of str/int, slice
        region : pyoti.region.region.Region
        resolution : float
            Actual resolution of data (and sections) returned by this
            Evaluator in datapoints per second.
        moving_filter : str
            The moving filter to be applied for `self.get_data()`. Defaults to
            'mean'.
        filter_time : float
            Desired time containing datapoints to be used with a running
            filter. If `filter_time` == 0.0, averaging is switched off. If
            `filter_time` > 0.0, signal is filtered according to the given
            `filter_time`. If `filter_time` < 0.0, signal is filtered according
            to `resolution` (`self.decimate`). The actual datapoints beeing
            used to filter the data can be checked with the property
            `self.window`.
        window : int
            Actual datapoints beeing used to filter the data.
        resolution_sf : float
            Actual resolution the SignalFeature works with.
        moving_filter_sf : str
            The moving filter to be applied to the signal for the
            SignaleFeature. Defaults to 'mean'.
        filter_time_sf : float
            The time for the moving filter, the signal for the SignalFeature
            should be filtered with. See attribute `filter_time` for details.
            Defaults to `filter_time`. The actual datapoints beeing used to
            filter the signal for the SignalFeature can be checked with the
            property `self.window_sf`.
        window_sf : int
            Actual datapoints beeing used to filter the signal for the
            SignalFeature.
        timevector
        decimate
        decimate_sf
        calibration
        """
        # resolution for (almost) all properties, e.g. displacement, force, ...
        # self.update() once only, by setting self.region = region!
        # resolution 1.000 Hz, samplingrate 40.000 Hz -> 40 points

        self._resolution = resolution or region.samplingrate  # Hz
        self.moving_filter = 'mean'
        self._filter_time = filter_time  # s
        self._resolution_sf = resolution_sf or self._resolution
        self.moving_filter_sf = 'mean'
        self._filter_time_sf = filter_time_sf or filter_time

        self._sf_class = sf_class or SignalFeature
        self._sf = None

        self.traces_sf = traces_sf
        self.region = region  # implicitly calls self.update()

        self.calibration = calibration

        self.rfigure = None

    def get_data(self, traces=None, samples=None, moving_filter=None,
                 filter_time=-1.0, copy=False, time=False, resolution=0):
        """
        Get data of `self.region` with a defined resolution and smoothed by
        a moving filter.

        Parameters
        ----------
        traces : str or list of str, optional
            see parameters of method `pyoti.region.Region.get_data()`
        samples : int or list of int or slice, optional
            Index of samples that should be returned. The index is relative to
            the data returned by this method without any parameter given. If
            you want to have a defined resolution, via the paramters
            `resolution` or the attribute `self.resolution`, do only use int or
            slice and do set the step value of the slice to None or 1.
        moving_filter : str, optional
            The moving filter to apply to the data. Can be one of 'mean' or
            'median'. Only applied, if `filter_time` != 0.0. Defaults to
            `self.moving_filter`.
        filter_time : float, optional
            Length of the window in s used to smooth the data with a moving
            filter. If `filter_time` is < 0.0 (default), use
            `self.filter_time` as the window length. If `filter_time` == 0.0,
            do not smooth the data.
        copy: bool, optional
            see parameters of method `pyoti.region.Region.get_data()`
        time : bool, optional
            Add the time as the first trace.
        resolution : int, optional
            Set the resolution of the returned data. If `resolution` <= 0
            (default), use `self.resolution` to set the resolution of the data.

        Returns
        -------
        2D numpy.ndarray
        """
        if resolution <= 0:
            decimate = self.decimate
        else:
            decimate = self._get_decimate(resolution)

        moving_filter = moving_filter or self.moving_filter

        if filter_time < 0.0:
            window = self.window
        elif filter_time > 0.0:
            window = max(1, int(filter_time * self.region.samplingrate))
        else:
            # filter_time == 0.0 # no averaging
            window = 1

        # Convert samples into the scope of self.region.get_data()
        samples = self.decimate_and_limit(samples)

        return self.region.get_data(traces=traces, samples=samples,
                                    moving_filter=moving_filter, window=window,
                                    decimate=decimate, copy=copy, time=time)

    def update(self):
        window = self.window_sf
        decimate = self.decimate_sf
        resolution = self.resolution_sf
        moving_filter = self.moving_filter_sf

        signals = self.region.get_data(self.traces_sf, copy=False,
                                       moving_filter=moving_filter,
                                       window=window, decimate=decimate)

        if self._sf is None:
            self._sf = self._sf_class(signals, resolution=resolution)
        else:
            self._sf.update(signals, resolution)

    def sections(self, slices=True, decimate=None, range_concat=False,
                 info=False, **kwargs):
        """
        Parameters
        ----------
        slices : bool, optional
            Return slices in a list (True) or a 2D numpy.ndarray with
            start/stop pairs in rows (False). Only evaluated, if `range_concat`
            is False.
        decimate : int, optional
            The step value for the slices returned. Only evaluated, if `slices`
            is True and `range_concat` is False.
        range_concat : bool, optional
            Return one 1D numpy.ndarray with indices, instead of individual
            slices or segments.
        info : bool, optional
            Additionally to the sections, return a 2D numpy.ndarray, containing
            infos about the individual sections (like which axis), if `info` is
            implemented in `self._sections()`. For details about the parmater
            see `self._sections()`.
        **kwargs, optional
            See method `self._sections()` for possible parameters.
        """
        # call function usually overwritten by subclasses
        sections = self._sections(info=info, **kwargs)
        if info and isinstance(sections, tuple):
            infos = sections[1]
            sections = sections[0]

        # Sort sections according to first index
        if sections.ndim == 1:
            sort = sections.argsort(axis=0)
        else:
            sort = sections.argsort(axis=0)[:, 0]
        sections = sections[sort]

        if info:
            infos = infos[sort]

        # undecimate and limit the sections
        sections = self.undecimate_and_limit(sections)

        # Convert segments (only, no idx!) to slices
        if slices and not range_concat and sections.ndim >= 2:
            sections = sn.idx_segments_to_slices(sections, decimate=decimate)

        # Convert segments (only, no idx!) to concatenated ranges
        if range_concat and sections.ndim >= 2:
            sections = sn.idx_segments_to_range(sections)

        if info:
            return sections, infos
        return sections

    def undecimate_and_limit(self, sections):
        """
        Convert the indices given with the resolution of `self.resolution_sf`
        to indices given with the resolution of `self.resolution`.
        """
        # undecimate the sections
        undecimate = self.decimate_sf / self.decimate
        max_stop = int(np.ceil(self.region.datapoints / self.decimate))
        sections = np.round(sections * undecimate).astype(int)
        sections = sn.limit_segments(sections, min_start=0, max_stop=max_stop)
        return sections

    def decimate_and_limit(self, samples):
        """
        Convert the indices given with the resolution of `self.resolution`
        to indices given with the resolution of `self.region.samplingrate`.
        """
        if samples is None:
            return None

        decimate = self.decimate
        max_stop = self.region.datapoints

        # Convert different samples types into nunmpy.ndarray, which can be
        # checked by the function `sn.limit_segments`.
        if isinstance(samples, int):
            _samples = np.array([samples])
        elif isinstance(samples, slice):
            _samples = np.array([[samples.start or 0,
                                 samples.stop or self.datapoints]])
            step = samples.step
            if step is not None:
                step = np.round(samples.step * decimate).astype(int)
        else:
            _samples = samples
        _samples = np.round(_samples * decimate).astype(int)
        _samples = sn.limit_segments(_samples, min_start=0, max_stop=max_stop)

        if isinstance(samples, slice):
            samples = slice(_samples[0, 0], _samples[0, 1], step)
        else:
            samples = _samples

        return samples

    def _sections(self, **kwargs):
        sections = np.empty((0, 2), dtype=int)
        # create sections (segments or idx)
        return sections

    def init_rfig(self, legend=True, show=True, plot_params=None):
        if self.rfigure is not None:
            plt.close(self.rfigure)

        # Initialize proper plot parameters
        gui.set_plot_params(plot_params=plot_params)

        # create static figure
        self.rfigure = self._rfigure(legend=legend)

        if show:
            self.rfigure.show()

    def _rfigure(self, legend=True):
        figure = None
        # create the figure ...
        return figure

    @property
    def region(self):
        return self._region

    @region.setter
    def region(self, region):
        if region is None:
            raise TypeError("Missing required argument: 'region'")
        self._region = region
        self.update()

    @property
    def resolution(self):
        resolution = self.region.samplingrate / self.decimate
        return resolution

    @resolution.setter
    def resolution(self, resolution):
        self._resolution = resolution
        self.update()

    @property
    def filter_time(self):
        return self._filter_time

    @filter_time.setter
    def filter_time(self, filter_time):
        self._filter_time = filter_time
        self.update()

    @property
    def window(self):
        if self.filter_time < 0.0:
            window = self.decimate
        elif self.filter_time > 0.0:
            window = max(1, int(self.filter_time * self.region.samplingrate))
        else:  # filter_time == 0.0 # no averaging
            window = 1
        return window

    @property
    def resolution_sf(self):
        resolution_sf = self.region.samplingrate / self.decimate_sf
        return resolution_sf

    @resolution_sf.setter
    def resolution_sf(self, resolution_sf):
        self._resolution_sf = resolution_sf
        self.update()

    @property
    def filter_time_sf(self):
        return self._filter_time_sf

    @filter_time_sf.setter
    def filter_time_sf(self, filter_time_sf):
        self._filter_time_sf = filter_time_sf
        self.update()

    @property
    def window_sf(self):
        if self.filter_time_sf < 0.0:
            window_sf = self._get_decimate(self._resolution_sf)
        elif self.filter_time_sf > 0.0:
            window_sf = max(1, int(self.filter_time_sf
                                   * self.region.samplingrate))
        else:  # window_sf == 0.0 # no averaging
            window_sf = 1
        return window_sf

    @property
    def timevector(self):
        return self.region.timevector[::self.decimate]

    @property
    def decimate(self):
        return self._get_decimate(self._resolution)

    @property
    def datapoints(self):
        return int(np.ceil(self.region.datapoints / self.decimate))

    @property
    def decimate_sf(self):
        return self._get_decimate(self._resolution_sf)

    def _get_decimate(self, resolution):
        return max(1, int(np.round(self.region.samplingrate
                                   / max(1, resolution))))

    @property
    def calibration(self):
        if hasattr(self, '_calibration') and self._calibration is not None:
            return self._calibration
        else:
            return self.region.calibration

    @calibration.setter
    def calibration(self, calibration):
        self._calibration = calibration



# -*- coding: utf-8 -*-
"""
Created on Wed Mar  9 17:07:32 2016

@author: Tobias Jachowski
"""
import matplotlib.pyplot as plt
import numpy as np
import time
from matplotlib.widgets import SpanSelector

from . import traces as tc


class GRS(object):
    """
    The GRS (Graphical Region Selector) creates a user interface to show data
    and select a range of the data.

    Attributes
    ----------
    ifigure : matplotlib.figure.Figure
        The current figure.
    displayrate : float
        Resolution of the plotted data in Hz.
    """

    def __init__(self, **kwargs):
        self._dpi = 100.0  # pixel per inch
        self._width = 800  # pixel per column
        self._height = 150  # pixel per row
        self._left = 65  # pixel for left
        self._right = 20  # pixel for right
        self._bottom = 55  # pixel for bottom
        self._top = 55  # pixel for top
        self._hspace = 18  # pixel between rows
        self._wspace = 18  # pixel between columns
        self.ifigure = None
        # _axspan is used by adjust_region AND _set_time_span, list of
        # indicators, which span is going to be used for every section
        self._axspan = {}
        # self._ax_OK = None
        self._tmin = 0
        self._tmax = 0
        self._minspan = 0.1  # s
        self._pre_title = "Select timespan (leftclick, draw, release)"
        self.displayrate = 1000.0  # display max 1000.0 datapoints per second
        self._onselect_callbacks = []
        self._spanselector = {}
        self._axspan = {}
        self.select = True

    def add_onselect_callback(self, function=None):
        """
        Register a function as a callback, when the selection of the data is
        changed.

        Parameters
        ----------
        function : function or method
            The function needs to accept two parameters tmin and tmax

        Returns
        -------
        bool
            True, if function was registered or is already registered as a
            callback function. Otherwise False.
        """
        if function is not None and function not in self._onselect_callbacks:
            self._onselect_callbacks.append(function)
            return True
        return False

    def set_onselect_callback(self, function=None):
        """
        Set a function as a callback, when the selection of the data is
        changed and deregister all other callback functions.

        Parameters
        ----------
        function : function or method
            The function needs to accept two parameters tmin and tmax.

        Returns
        -------
        bool
            True, if function was set as the callback function. Otherwise
            False.
        """
        if function is not None:
            self._onselect_callbacks.clear()
            self._onselect_callbacks.append(function)
            return True
        return False

    def _ifigure(self, timevector, data, samplingrate, traces, tmin=None,
                 tmax=None, xlim=None, description=None):
        """
        Create a figure based on the given timevector and data that lets the
        user choose a timespan (tmin, tmax):
          - Create figure with axes for all traces
          - Add SpanSelector for selecting timespan and set initial tmin and
            tmax according to given region
          - return figure

        Parameters
        ----------
        timevector : np.array
            1D array with the timevector to plot
        data : np.array
            The data array containing the traces to plot
        samplingrate : float
            The samplingrate of the data in Hz
        traces : list of str
            List with names of traces to use for data, COLOR and NAME
        tmin : float, optional
            tmin for the span selector
        tmax : float, optional
            tmax for the span selector
        xlim : tuple, optional
            tuple with (tmin, tmax) to set the range of the data to plot
        description : str, optional
            str describing the purpose of the region to select. Is used as a
            heading

        Returns
        -------
        matplotlib.figure.Figure
        """

        traces = tc.normalize(traces)

        rows = len(traces)
        columns = 1

        # Create figure and axes
        ifigure, axes_ = plt.subplots(rows, columns, sharex=True)
        if rows * columns == 1:
            axes_ = np.array([axes_])
        ax = dict(list(zip(traces, axes_)))

        if tmin is None:
            tmin = timevector[0]
        if tmax is None:
            tmax = timevector[-1]
        self._tmin = tmin
        self._tmax = tmax

        if xlim is None:
            xlim = (timevector[0], timevector[-1])

        decimate = max(1, int(np.round(samplingrate / self.displayrate)))

        # plot traces
        for i, trace in enumerate(traces):
            ax[trace].plot(timevector[::decimate],
                           data[:, i][::decimate] * tc.factor(trace),
                           tc.color(trace))
            ax[trace].set_ylabel(tc.label(trace))
            ax[trace].ticklabel_format(useOffset=False)
            ax[trace].set_xlim(*xlim)
            ax[trace].grid(True)

            if self.select:
                # indicators, which span is going to be used
                self._axspan[trace] = ax[trace].axvspan(self._tmin,
                                                        self._tmax,
                                                        facecolor='y',
                                                        alpha=0.1)
                # selectors for timespan to be used
                self._spanselector[trace] = SpanSelector(ax[trace],
                                                         self._set_time_span,
                                                         'horizontal',
                                                         useblit=True,
                                                         minspan=self._minspan)

        # Add OK button
        # TODO: improve scaling of OK button size
        # self._ax_OK = ifigure.add_axes([ 0.9, 0.01, 0.06, 0.03])
        # button = plt.Button(self._ax_OK, 'OK')
        # def ap(event):
        #     self.close_ifig()
        # button.on_clicked(ap)

        ax[traces[-1]].set_xlabel('Time (s)')

        if description is None:
            ifigure.suptitle(self._pre_title)
        else:
            ifigure.suptitle(self._pre_title + ' for ' + description)

        return ifigure

    def update_ifig(self, timevector, data, samplingrate, traces, xlim=None):
        """
        Update the data of self.ifigure.

        Parameters
        ----------
        timevector : array
            The timeline of the data.
        data : 2D array
            The data to be plotted.
        samplingrate : float
            The samplingrate of the data in Hz. Data needs to have as many
            traces as the data that is currently plotted in `self.ifigure`.
        traces : str or list of str
            The traces contained in the data. `traces` has to contain the same
            traces as the data that is currently plotted in `self.ifigure`.
        xlim : tuple, optional
            Set the range (xmin, xmax) of the data to be plotted in units of
            timevector.
        """
        traces = tc.normalize(traces)
        ax = dict(list(zip(traces, self.ifigure.axes)))

        decimate = max(1, int(np.round(samplingrate / self.displayrate)))

        if xlim is None:
            xlim = (timevector[0], timevector[-1])

        # plot traces
        for i, trace in enumerate(traces):
            ax[trace].lines[0].set_data(
                timevector[::decimate],
                data[:, i][::decimate] * tc.factor(trace))
            ax[trace].set_xlim(*xlim)

        self.ifigure.canvas.draw()

    def close_ifig(self):
        """
        Close the figure the user has selected the region from and deregister
        all callback functions.
        """
        # remove references to SpanSelectors, to free memory
        self._axspan.clear()
        self._spanselector.clear()

        # remove onselect callbacks
        self._onselect_callbacks.clear()

        if self.ifigure is not None:
            # Remove button 'OK'
            # self._ax_OK.clear()
            # self._ax_OK.set_axis_off()
            # self._ax_OK = None

            # Some matplotlib backends will throw an error when trying to draw
            # the canvas. Simply ignoring the error that could happen here will
            # prevent the figure from not beeing closed, left open, and
            # preventing the next figure to be drawn.
            # Even though the except: pass clause is considered bad, here the
            # worst thing that could happen, is that the static figure produced
            # by the matplotlib backend upon closing is not up to date.
            # Therefore, except: pass should be considered as an acceptable
            # workaround for this case.
            try:
                # force redraw of the figure to remove the button in the nbagg
                # backend for the static figure
                self.ifigure.canvas.draw()

                # give backend time to redraw the figure (hack) before closing
                time.sleep(1)
            except:
                pass

            # close the figure
            plt.close(self.ifigure)

        self.ifigure = None

    def init_ifig(self, timevector, data, samplingrate, traces, tmin=None,
                  tmax=None, onselect_cb=None, xlim=None, description=None,
                  select=True):
        """
        Create and show a figure based on the given timevector and data that
        lets the user choose a timespan (tmin, tmax). Whenever the user changes
        the timespan, the function onselect_cb is called with parameters tmin
        and tmax.

        Parameters
        ----------
        timevector : np.array
            1D array with the timevector to plot
        data : np.array
            The data array containing the traces to plot
        samplingrate : float
            The samplingrate of the data in Hz
        traces : list of str
            List with names of traces to use for data, COLOR and NAME
        tmin : float, optional
            tmin for the span selector
        tmax : float, optional
            tmax for the span selector
        onselect_cb : function, optional
            A function which is called whenever the timespan was changed by the
            user. onselect_cb() is called with the two parameters tmin and
            tmax: onselect_cb(tmin, tmax).
        xlim : tuple, optional
            tuple with (tmin, tmax) to set the range of the data to plot.
            Defaults to (timevector[0], timevector[-1]).
        description : str, optional
            str describing the purpose of the region to select. Is used as a
            heading for the figure.
        """
        self.select = select

        # nbagg backend needs to have the figure closed and recreated
        # whenever the code of the cell displaying the figure is executed.
        # A simple update of the figure would let it disappear. Even a
        # self.ifigure.show() wouldn't work anymore.
        # For backends this just means a bit of extra calculation.
        # Therefore, close the figure first before replotting it.
        self.close_ifig()

        # replace old onselect_callback with new one
        self.set_onselect_callback(onselect_cb)

        # set plot parameters
        traces = tc.normalize(traces)
        plot_params = self._calculate_plot_params(len(traces), 1)
        set_plot_params(plot_params)

        # create interactive figure
        self.ifigure = self._ifigure(timevector,
                                     data,
                                     samplingrate,
                                     traces,
                                     tmin=tmin,
                                     tmax=tmax,
                                     xlim=xlim,
                                     description=description)

        # show the interactive figure
        self.ifigure.show()

    def _calculate_plot_params(self, rows, columns):
        """
        Calculate proper figure geometry and return matplotlib parameters
        accordingly. The geometry is calculated depending on the number of rows
        and columns.

        Parameters
        ----------
        rows : int
            Number of rows to plot data.
        columns : int
            Number of columns to plot data.

        Returns
        -------
        dict
            Matplotlib parameters. Can by set with `self.set_plot_params()`.
        """
        dpi = self._dpi

        fig_pixel_width  = columns * float(self._width)
        fig_pixel_height = rows    * float(self._height)

        left   =        self._left   / fig_pixel_width
        right  = 1.00 - self._right  / fig_pixel_width
        bottom =        self._bottom / fig_pixel_height
        top    = 1.00 - self._top    / fig_pixel_height

        wspace = self._wspace / float(self._width)
        hspace = self._hspace / float(self._height)

        plot_params = {
            'dpi': dpi,
            'fig_pixel_width': fig_pixel_width,
            'fig_pixel_height': fig_pixel_height,
            'left': left,
            'right': right,
            'bottom': bottom,
            'top': top,
            'wspace': wspace,
            'hspace': hspace
        }

        return plot_params

    def _set_time_span(self, tmin, tmax):
        """
        Set the timespan according to the SpanSelector and update the figure
        accordingly. Call all registered callback functions.

        This function is called upon any change of the SpanSelector.

        Parameters
        ----------
        tmin : float
            Start time point in s.
        tmax : float
            End time point in s.
        """
        self._tmin = tmin
        self._tmax = tmax
        for ax in list(self._axspan.values()):
                    ax.set_xy([[tmin, 0],  # lower left corner
                               [tmin, 1],  # upper left corner
                               [tmax, 1],  # upper right corner
                               [tmax, 0],  # lower right corner
                               [tmin, 0]])  # lower left corner
        self.ifigure.canvas.draw()
        self._fire_onselect_callbacks()

    def _fire_onselect_callbacks(self):
        """
        Call all registered callback functions.
        """
        # call onselect callback functions
        for function in self._onselect_callbacks:
            function(self._tmin, self._tmax)


def set_plot_params(plot_params=None):
    """
    Set matplotlib parameters.

    Parameters
    ----------
    plot_params : dict, optional
        Dictionary of matplotlib parameters to be set. The following default
        parameters are used, if not existent in plot_params:
        {
            'dpi' : 100.0,
            'fig_pixel_width' : 800.0,
            'fig_pixel_height' : 600.0,
            'left' : 65 / fig_pixel_width,
            'right' : 1.00 - 20 / fig_pixel_width,
            'bottom' : 55 / fig_pixel_height,
            'top' : 1.00 - 55 / fig_pixel_height,
            'wspace' : 18 / float(fig_pixel_width),
            'hspace' : 18 / float(fig_pixel_height),
            'lines.linewidth' : 0.25,
            'figure.facecolor' : '1.00',
            'figure.edgecolor' : '1.00',
            'axes.linewidth' : 1,
            'axes.labelsize' : 9,
            'legend.fontsize' : 9,
            'ytick.labelsize' : 7,
            'xtick.labelsize' : 7
        }
    """
    if plot_params is None:
        plot_params = {}

    dpi = plot_params.pop('dpi', 100.0)

    fig_pixel_width = plot_params.pop('fig_pixel_width', 800.0)
    fig_pixel_height = plot_params.pop('fig_pixel_height', 600.0)

    fig_width  = fig_pixel_width  / dpi
    fig_height = fig_pixel_height / dpi

    left   = plot_params.pop('left',         65 / fig_pixel_width)
    right  = plot_params.pop('right', 1.00 - 20 / fig_pixel_width)
    bottom = plot_params.pop('bottom',       55 / fig_pixel_height)
    top    = plot_params.pop('top',   1.00 - 55 / fig_pixel_height)

    wspace = plot_params.pop('wspace', 18 / float(fig_pixel_width))
    hspace = plot_params.pop('hspace', 18 / float(fig_pixel_height))

    lines_linewidth = plot_params.pop('lines.linewidth', 0.25)

    figure_facecolor = plot_params.pop('figure.facecolor', '1.00')
    figure_edgecolor = plot_params.pop('figure.edgecolor', '1.00')

    axes_linewidth = plot_params.pop('axes.linewidth', 1)

    axes_labelsize = plot_params.pop('axes.labelsize', 9)
    legend_fontsize = plot_params.pop('legend.fontsize', 9)
    ytick_labelsize = plot_params.pop('ytick.labelsize', 7)
    xtick_labelsize = plot_params.pop('xtick.labelsize', 7)

    rcparams = plt.rcParams.copy()
    rcparams = {
        'figure.figsize': (fig_width, fig_height),
        'figure.dpi': dpi, 'savefig.dpi': dpi,
        'figure.subplot.left': left,
        'figure.subplot.right': right,
        'figure.subplot.bottom': bottom,
        'figure.subplot.top': top,
        'figure.subplot.wspace': wspace,
        'figure.subplot.hspace': hspace,
        'figure.facecolor': figure_facecolor,
        'figure.edgecolor': figure_edgecolor,
        'lines.linewidth': lines_linewidth,
        'axes.linewidth': axes_linewidth,
        'axes.labelsize': axes_labelsize,
        'legend.fontsize': legend_fontsize,
        'ytick.labelsize': ytick_labelsize,
        'xtick.labelsize': xtick_labelsize
        }

    plt.rcParams.update(rcparams)



# -*- coding: utf-8 -*-
"""
Created on Thu Mar 10 19:13:51 2016

@author: Tobias Jachowski
"""
import os
import persistent
from functools import wraps

from .. import helpers as hp


def if_needs_file(func):
    @wraps(func)
    def func_wrapper(self, *args, **kwargs):
        if self._needs_file:
            return func(self, *args, **kwargs)
        print("DataSource does not need files to get data!")
    return func_wrapper


class DataSource(persistent.Persistent):
    name = 'Generic data source'
    samplingrate = 1000  # default 1000 Hz
    _needs_file = False

    def __init__(self, filename, directory=None, root=None, **kwargs):
        if filename is not None:
            self._needs_file = True
            self._root = root
            filename, absdir, fullfilename = hp.file_and_dir(filename,
                                                             directory)
            self._filename = filename
            self._directory = os.path.relpath(absdir, self.rootdir)
            self._filename_orig = filename
            self._directory_orig = self._directory
            self._absdir_orig = absdir

    def as_array():
        """
        This function must return a 2D numpy.ndarray, containing the data.
        """
        return None

    @property
    @if_needs_file
    def rootdir(self):
        """
        The directory of the experiment file.
        """
        if self._root is None:
            print("The datasource is not stored in an experiment file!\n"
                  "Therefore, no directory is known!\n"
                  "Instead, the current working directory is used!")
            return os.path.realpath('.')
        return self._root._v_absdir

    @property
    @if_needs_file
    def directory(self):
        """
        The directory of the data file relative to the experiment file.
        """
        return self._directory

    @property
    @if_needs_file
    def absdir(self):
        """
        The absolute directory of the data file.
        """
        return os.path.realpath(os.path.join(self.rootdir, self.directory))

    @property
    @if_needs_file
    def filename(self):
        """
        The filename of the data file.
        """
        return self._filename

    @property
    @if_needs_file
    def absfile(self):
        """
        The absolute path to the data file.
        """
        return os.path.join(self.absdir, self.filename)

    @property
    @if_needs_file
    def directory_orig(self):
        """
        The relative directory the data file was originally loaded from.
        """
        return self._directory_orig

    @property
    @if_needs_file
    def filename_orig(self):
        """
        The original name of the data file.
        """
        return self._filename_orig

    @property
    @if_needs_file
    def absdir_orig(self):
        """
        The absolute directory the data file was loaded from.
        """
        return self._absdir_orig

    @property
    @if_needs_file
    def absfile_orig(self):
        """
        The absolute path to the data file it was originally loaded from.
        """
        return os.path.join(self.absdir_orig, self.filename_orig)



# -*- coding: utf-8 -*-
# """
# - Author: steve simmert
# - E-mail: steve.simmert@uni-tuebingen.de
# - Copyright: 2015
# """

import matplotlib.gridspec as gridspec
import matplotlib.pyplot as plt

col_dict = {'x': 'blue',
            'y': 'green',
            'z': 'orange',
            'psdX': 'blue',
            'psdY': 'green',
            'psdZ': 'orange',
            'drag': 'DarkTurquoise',
            'ox': 'LightSteelBlue',
            'oy': 'DarkSeaGreen',
            'oz': 'BlanchedAlmond',
            'opsdX': 'LightSteelBlue',
            'opsdY': 'DarkSeaGreen',
            'opsdZ': 'BlanchedAlmond',
            'odrag': 'LightCyan',
            0: 'blue',
            1: 'green',
            2: 'orange',
            3: 'DarkTurquoise',
            4: 'MediumOrchid',
            5: 'LimeGreen',
            6: 'DeepSkyBlue',
            7: 'Teal',
            8: 'Olive',
            9: 'Tomato',
            10: 'LightSteelBlue',
            11: 'DarkSeaGreen',
            12: 'BurlyWood',
            13: 'LightCyan',
            14: 'Plum',
            15: 'LightGreen',
            16: 'LightBlue',
            17: 'PowderBlue',
            18: 'DarkKhaki',
            19: 'LightCoral'
            }


def set_axis_font_size(axis, font_size):
    """ Set the font size of the title, axis labels and tick labels. """
    for item in ([axis.title, axis.xaxis.label, axis.yaxis.label] +
                 axis.get_xticklabels() + axis.get_yticklabels()):
        item.set_fontsize(font_size)


def set_sym_y_labels(axis, mean=None):
    """
    Set symmetric y-limits for the specified axis.
    """
    low, upp = axis.get_ylim()

    if mean is None:
        mean = (upp - low) / 2

    if mean - low <= upp - mean:
        upp_ = upp
        low_ = mean - (upp - mean)
    else:
        upp_ = mean + (mean - low)
        low_ = low

    axis.set_ylim(low_, upp_)


def add_plot_to_figure(figure,
                       xdata,
                       ydata,
                       xerr=[],
                       yerr=[],
                       subplot=(1, 1, 1),
                       axis=None,
                       sharex=None,
                       sharey=None,
                       fmt='-',
                       fontsize=16,
                       markersize=8,
                       linewidth=1.5,
                       alpha=1.0,
                       label='',
                       xlabel='',
                       ylabel='',
                       title='',
                       nbins=4,
                       logplot= False,
                       showLegend=False,
                       figsize=(9, 6),
                       legend_kwargs={},
                       **plot_kwargs):
    """
    Adds a subplot to an existing figure and returns the added axis.

    The function discriminates between normal plots and plots with error bars.
    Errorbars are used if xerr AND yerr OR yerr are not empty arrays.

    Arguments
    ---------
    figure : matplotlib.pyplot.figure or None
    xdata : array
        x data to be plotted.
    ydata : array
        y data to be plotted.
    xerr : array or list
        errors for the x data. Default is an empty list.
    yerr : array or list
        errors for the y data. Default is an empty list.
    subplot : tuple
        Three-element tuple defining the subplot to be used. Defaults to
        (1,1,1)
    axis : matplotlib.Axis or None
        Axis that is used for the plot. If None a new one is added.
    sharex : matplotlib.Axis or None
        shared x axis to be used
    sharey : matplotlib.Axis or None
        shared y axis to be used
    fmt : str
        format string: e.g. '+b'
    fontsize : int
    markersize : int
    linewidth : float
    alpha : float
        transparency for the line/marker
    xlabel : str
    ylabel : str
    title : str
    nbins : int
        number of bins for the axes ticks and labels
    showLegend : bool
        If True, also plots the legend
    legend_kwargs : dict
        dict of legend arguments, e.g. a = {'bbox_to_anchor': (1.05, 1),
        'loc': 2, 'borderaxespad': 0}
    **plot_kwargs : keyword arguments
        keyword arguments that are handed over to pyplot.plot()
        or errorbar function, respectively
    """
    if figure is None:
        figure = plt.figure(figsize=figsize)

    if axis is not None:
        ax = axis
    else:
        ax = figure.add_subplot(*subplot, sharex=sharex, sharey=sharey)

    if len(xerr) == 0 and len(yerr) == 0:
        ax.plot(xdata, ydata, fmt,
                markersize=markersize, linewidth=linewidth, label=label,
                alpha=alpha, **plot_kwargs)
    elif ((xerr == [] or sum(xerr) == 0) and len(yerr) > 0):
        ax.errorbar(xdata, ydata, yerr=yerr, fmt=fmt,
                    markersize=markersize, linewidth=linewidth, label=label,
                    alpha=alpha, **plot_kwargs)
    else:
        ax.errorbar(xdata, ydata, xerr=xerr, yerr=yerr, fmt=fmt,
                    markersize=markersize, linewidth=linewidth, label=label,
                    alpha=alpha, **plot_kwargs)
    if logplot:
        ax.set_xscale('log')
        ax.set_yscale('log')
    else:
        ax.set_xscale('linear')
        ax.set_yscale('linear')

    if showLegend:
        if (legend_kwargs != {}):
            ax.legend(**legend_kwargs)
        else:
            ax.legend()

    if ax.get_xscale() is not 'log' and ax.get_yscale() is not 'log':
        ax.locator_params(nbins=nbins)
        ax.xaxis.get_major_formatter().set_powerlimits((-3, 3))
        ax.yaxis.get_major_formatter().set_powerlimits((-3, 3))

    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.set_title(title)

    set_axis_font_size(ax, fontsize)

    return ax


def get_gridded_axis(figure=None,
                     figsize=(9, 6),
                     grid=(1, 1),
                     axslice=(slice(None), slice(None)),
                     **axes_kwargs):
    """
    Return an axis of a grid.

    Arguments
    ---------
    figure : matplotlib.Figure
    figsize : tuple
        Size of the figure in inch.
    grid : tuple
        Size of the grid, e.g. grid=(9, 7) for a grid of 9 rows and 7 columns.
    axslice : tuple(row_slice, col_slice)
        Tuple of slices that define the rows and cols to be used for the axis
        that shall be returned. E.g. axslice=(slice(0,4,1), slice(3,6,1))
        for an axis expanding from row 0 to 3 and columns 3 to 5.
    **axes_kwargs : axis keywordarguments
    """
    if figure is None:
        fig = plt.figure(figsize=figsize)
    else:
        fig = figure
    gs = gridspec.GridSpec(*grid)
    ax = fig.add_subplot(gs[axslice], **axes_kwargs)
    return ax


def get_residual_plot_axes(nrows=7, ncols=1,
                           row_lim=5, figure=None, figsize=(9, 6)):
    """
    Generate an axes grid in a figure and return a list of 2-tuples with
    the upper axis and lower axis.

    Arguments
    ---------
    nrows, ncols : int
        Number of rows and columns of the axes-grid.
    row_lim : int
        Row from which the lower axis starts.
    figure : Figure
    figsize : tuple(width, height) in inches
    """
    if figure is None:
        fig = plt.figure(figsize=figsize)
    else:
        fig = figure

    grid = (nrows, ncols)

    axes = []

    for col in range(ncols):
        ax1 = get_gridded_axis(figure=fig,
                               figsize=figsize,
                               grid=grid,
                               axslice=(slice(0, row_lim, 1),
                                        slice(col, col + 1, 1)))
        if row_lim < nrows:
            ax2 = get_gridded_axis(figure=fig,
                                   figsize=figsize,
                                   grid=grid,
                                   axslice=(slice(row_lim, nrows, 1),
                                            slice(col, col + 1, 1)))
        else:
            ax2 = None

        axes.append((ax1, ax2))

    return axes



#!/usr/bin/env python3

from setuptools import setup, find_packages, findall
#from distutils.core import setup
import os
import re
import importlib


# read the version from VERSION.txt
with open('VERSION.txt') as version_file:
    __version__ = version_file.read().strip()

# write VERSION.txt files in both packages
for pkg in ['pyoti', 'pyotc']:
    ptf = os.path.join(*['.', pkg, 'VERSION.txt'])
    with open(ptf, 'w') as vf:
        vf.write('{}\n'.format(__version__))
    print('created version file: {}'.format(ptf))

# get the required packages from requirements.txt and check if they are already
# installed. Probably obsolete for newer setuptools versions when using required
# packages arguments of setup() function.
with open('requirements.txt', 'r') as rfile:
    requs = [re.split('([\s=<>]+)', l.replace('\n', '').strip(' '))
             for l in rfile.readlines()
             if not(l.startswith('#')) and l != '\n']

abort = False

for requ in requs:
    try:
        importlib.import_module(requ[0])
    except:
        abort = True
        print("Module {} not available.".format(requ[0]))
        if len(requ) > 1:
           print("\tPyOTIC requires version {}".format(''.join(requ[1:]).strip()))

if abort:
    print("Try to install the missing package(s) first, e.g. type:\n"
          "$ pip(3) install {}".format(' '.join(l[0] for l in requs)))
        
if not abort:
    long_desc = """
                Time-dependent signal analysis and calibration of optical
                tweezers by power spectral denstity analysis.
                """
    license = 'Apache-2.0'

    setup(name='pyotic',
          version=__version__,
          author='Tobias Jachowski and Steve Simmert',
          author_email='steve.simmert@uni-tuebingen.de',
          url='https://github.com/cellular-nanoscience/pyotic',
          license=license,
          description="Time-dependent signal analysis and calibration of optical tweezers.",
          long_description = long_desc,
          platforms = ['Windows', 'Linux', 'Mac OS X'],
          classifiers=['Intended Audience :: Science/Research',
                       'Intended Audience :: Education',
                       'Framework :: Jupyter',
                       'License :: OSI Approved :: Apache Software License',
                       'Operating System :: OS Independent',
                       'Programming Language :: Python',
                       'Topic :: Scientific/Engineering :: Physics',
                       'Topic :: Scientific/Engineering :: Visualization', ],
          python_requires='>=3, <4',
          packages=find_packages(),
          package_dir = {'pyotc': 'pyotc',
                         'pyoti': 'pyoti'},
          package_data={'pyotc': ['VERSION.txt'],
                        'pyoti': ['VERSION.txt',
                                  './configs/calibration.cfg',
                                  './configs/groups.cfg',
                                  './configs/traces.cfg',
                                  './configs/record/GenericData.cfg',
                                  './configs/record/GenericDataFile.cfg',
                                  './configs/record.cfg',
                                  './plugins/modifications/generic.py',
                                  './plugins/modifications/rotation.py',
                                  './plugins/modifications/offset.py',
                                  './plugins/modifications/beadscan.py',
                                  './plugins/modifications/attachment.py',
                                  './plugins/modifications/touchdown.py',
                                  './plugins/modifications/baseline.py',
                                  './plugins/calibsources/cellnano.py',
                                  './plugins/calibsources/pyotic.py',
                                  './plugins/datasources/cellnano.py',
                                  './plugins/datasources/generic.py',
                                  './evaluate/fast.pyx',
                                  ]},
          )



# -*- coding: utf-8 -*-
"""
Created on Thu Mar 10 21:31:32 2016

@author: Tobias Jachowski
"""
import matplotlib.pyplot as plt
import numpy as np
from scipy.interpolate import UnivariateSpline

from pyoti.modification.modification import Modification, GraphicalMod
from pyoti import helpers as hp
from pyoti import traces as tc


class IBeadscan(GraphicalMod):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._ax = None

    def _figure(self):
        """
        Show the fitted model functions that are used to correct apply_data.
        """
        # create new figure and axes for adjusting the plateaus
        plot_params = {'lines.linewidth': 3.0}
        self._set_plot_params(plot_params)
        figure, ax_ = plt.subplots(1, 2, sharex=True, sharey=True)

        plottrace = self.modification.traces_apply
        ax = dict(list(zip(plottrace, ax_)))
        self._ax = ax

        # [ psdX, psdY ]
        traces = self.modification.traces_apply
        if 'psdZ' not in traces:
            traces.append('psdZ')
            traces.sort()

        # create lines to plot the corrections
        for trace in plottrace:
            ax[trace].plot([0], [0], c=tc.color(trace), marker='.', ls='')
            ax[trace].plot([0], [0], 'k-')
            ax[trace].plot([0], [0], 'y-', linewidth=3)
            ax[trace].legend(('fitdata', 'fit+extrapolation', 'fit'),
                             loc='best')
            ax[trace].set_xlabel(tc.label('psdZ'))
            ax[trace].set_ylabel(tc.label(trace))
            ax[trace].ticklabel_format(useOffset=False)
            ax[trace].grid(True)

        figure.suptitle("Crosstalk correction function for psdX and psdY from "
                        "psdZ.")

        return figure

    def _update_fig(self, **kwargs):
        plottrace = self.modification.traces_apply
        traces = self.modification.traces_apply
        traces.append('psdZ')
        Z = traces.index('psdZ')

        data_apply = self.modification._get_data_apply(
            traces=traces, decimate=True, copy=False)
        data_based = self.modification._get_data_based(
            traces=traces, decimate=True, copy=False)

        # Calculate the sorted indices for psdZ for plotting a "smooth" line
        # psdX, psdY, and psdZ sorted by increasing psdZ
        sortidx_apply = np.argsort(data_apply[:, Z])
        sortidx_based = np.argsort(data_based[:, Z])

        # get modification for traces psdX and psdY
        mod_index = hp.overlap_index(self.modification.traces_apply, plottrace)
        modification_apply \
            = self._modification(data_apply, self.samples_apply_decimated,
                                 traces, mod_index)
        modification_based \
            = self._modification(data_based, self.samples_based_decimated,
                                 traces, mod_index)

        ax = self._ax

        for trace in plottrace:
            lines = ax[trace].lines
            T = traces.index(trace)
            trace_idx_apply = self.modification.view_apply.traces_to_idx(trace)
            trace_idx_based = self.modification.view_based.traces_to_idx(trace)
            lines[0].set_data(data_based[:, Z], data_based[:, T])
            lines[1].set_data(data_apply[sortidx_apply, Z],
                              modification_apply[sortidx_apply,
                                                 trace_idx_apply])
            lines[2].set_data(data_based[sortidx_based, Z],
                              modification_based[sortidx_based,
                                                 trace_idx_based])

            # recompute ax.dataLim
            ax[trace].relim()
            # update ax.viewLim using new dataLim
            ax[trace].autoscale_view()

    def _modification(self, data, samples, data_traces, mod_index):
        """
        For compatibility reasons
        """
        psdZ_apply = self.modification._get_trace_data_apply(
            'psdZ', data, samples, data_traces, copy=False)

        # initialize the modification according to previously determined model
        # for X/Y
        mod_trace = hp.listify(mod_index)
        modification = np.zeros((len(data), len(mod_trace)))

        for i, trace in enumerate(mod_trace):
            modification[:, i] = self.modification._model[trace](psdZ_apply)

        return modification

    @property
    def samples_based_decimated(self):
        return slice(0, self.modification.view_based.datapoints,
                     self.modification.decimate)

    @property
    def samples_apply_decimated(self):
        return slice(0, self.modification.view_apply.datapoints,
                     self.modification.decimate)


class Beadscan(Modification):
    """
    Corrects for crosstalk from psdZ to psdX/psdY with a supplied beadscan
    in z axis.
    """
    GRAPHICALMOD = IBeadscan

    def __init__(self, **kwargs):
        traces_apply = ['psdX', 'psdY']
        super().__init__(datapoints=2500, traces_apply=traces_apply, **kwargs)

        # Define some attributes needed for the fitting of the beadscan
        # model for fitting psdX and psdY, usually UnivariatSpline
        self._model = {}

        # min and max value of psdZ used to determine the model
        self.minpsdZ = None
        self.maxpsdZ = None

    def _recalculate(self):
        # calculate data for fitting
        traces = self.traces_apply
        traces.append('psdZ')
        bin_means, bin_width = self.calculate_bin_means(traces=traces,
                                                        sorttrace=2)
        self.bin_means = bin_means

        # calculate the model for smoothing data
        self._calculate_model()

    def _print_info(self):
        for trace in self.traces_apply:
            print("    %s coeffs: %s"
                  % (tc.label(trace),
                     self._model[self.lia(trace)].get_coeffs()))

    def _modify(self, data, samples, data_traces, data_index, mod_index):
        psdZ_apply = self._get_trace_data_apply('psdZ', data, samples,
                                                data_traces, copy=False)

        # apply modification and return modified data according to previously
        # determined model for X/Y
        data_trace = hp.listify(data_index)
        mod_trace = hp.listify(mod_index)

        for dtrace, mtrace in zip(data_trace, mod_trace):
            data[:, dtrace] -= self._model[mtrace](psdZ_apply)

        return data

    def _get_trace_data_apply(self, trace, data, samples, data_traces,
                              copy=True):
        # get psdZ and reduce dimensionality for later calculation of
        # modification
        if trace in data_traces:  # trace already in data
            index = data_traces.index(trace)
            trace_apply = data[:, index]
        else:  # psdZ needs to be fetched extra
            trace_apply = self._get_data_apply(samples=samples, traces=trace,
                                               copy=False)[:, 0]  # [ trace ]

        # check for extrapolation limit of the previously calculated model
        minpsdZ = trace_apply.min()
        maxpsdZ = trace_apply.max()
        if minpsdZ < self.minpsdZ:
            if self.minpsdZ - minpsdZ > 0.01:
                print("Beadscan: Need to extrapolate the data for crosstalk "
                      "psdZ -> psdX/Y min by %.5f V"
                      % (self.minpsdZ - minpsdZ))
        if maxpsdZ > self.maxpsdZ:
            if maxpsdZ - self.maxpsdZ > 0.01:
                print("Beadscan: need to extrapolate the data for crosstalk "
                      "psdZ -> psdX/Y max by %.5f V"
                      % (maxpsdZ - self.maxpsdZ))

        if copy:
            return trace_apply.copy()
        else:
            return trace_apply

    def _calculate_model(self):
        """
        Use beadscan x (psdX, psdY) and y (psdZ) data to create a
        one-dimensional smoothing model fit.
        Be aware of the indexing X, Y and Z (defined above) -> NOT psdX, psdY,
        psdZ
        """
        X = 0  # psdX
        Y = 1  # psdY
        Z = 2  # psdZ

        self.minpsdZ = self.bin_means[:, Z].min()
        self.maxpsdZ = self.bin_means[:, Z].max()
        self._model[X] = UnivariateSpline(self.bin_means[:, Z],
                                          self.bin_means[:, X])
        self._model[Y] = UnivariateSpline(self.bin_means[:, Z],
                                          self.bin_means[:, Y])
        self.set_changed()

    @property
    def bin_means(self):
        if not hasattr(self, '_v_bin_means'):
            return None
        else:
            return self._v_bin_means

    @bin_means.setter
    def bin_means(self, bin_means):
        self._v_bin_means = bin_means


# The following is only to update to database version 0.8.0
class GBeadscan(Beadscan):
    pass



# -*- coding: utf-8 -*-
"""
Created on Tue May 15 12:32:48 2018

@author: Tobias Jachowski
"""
import h5py
import numpy as np
import os

from pyoti.data.datasource import DataSource


class HDF(DataSource):
    def __init__(self, filename, section=None, directory=None, datext='.bin',
                 parext='_para.dat', **kwargs):
        """
        parext : str, optional
            The extension of the parameter file ('_para.dat', default)
        """
        super().__init__(filename=filename, directory=directory, **kwargs)
        # TODO: Is it necessary to select a certain section?
        self.section = section

        self.samplingrate = get_samplingrate(self.absfile, section=section)

        self.name = ("HDF data originally loaded from \n"
                     "    %s with \n"
                     "    samplingrate %s Hz") % (self.absfile_orig,
                                                  self.samplingrate)

    def as_array(self):
        filename = self.absfile
        f = h5py.File(filename, "r")
        # TODO: for trace in ... do
            data = f[self.section]
        return data

    @property
    def parafile_orig(self):
        return self._parafile_orig

def get_samplingrate(filename, section=None):
    f = h5py.File(filename, "r")
    # TODO: howto get proper samplingrate
    samplingrate = f[section]["Sample rate (Hz)"]
    return samplingrate
